"""Athena Query Utilities"""
import json
import datetime
import itertools

def get_date_n_days_ago(n_days: int):
    """get the date n_days ago relative to now()"""
    return (
        datetime.datetime.now() - datetime.timedelta(days = n_days)
    ).strftime('%Y-%m-%d')

def get_query_parameters(
    sources: list,
    mediums: list,
    view_ids: list,
    excl_landing_pages: list,
    start_date: str,
    end_date: str
):
    if isinstance(sources, str):
        sources = json.loads(sources)
    if isinstance(mediums, str):
        mediums = json.loads(mediums)
    if isinstance(view_ids, str):
        view_ids = json.loads(view_ids)
    if isinstance(excl_landing_pages, str):
        excl_landing_pages = json.loads(excl_landing_pages)

    if start_date.isdigit():
        start_date = get_date_n_days_ago(int(start_date))

    if end_date.isdigit():
        end_date = get_date_n_days_ago(int(end_date))

    return (
        sources,
        mediums,
        view_ids,
        excl_landing_pages,
        start_date,
        end_date
    )

def encode_list_as_list_of_strings(items: list):
    """
        Encodes a list as a String list of items for Athena query.
        
        Adding a list of items to athena query argument requires it to be
        encoded as a string encoded list
    """
    items = ["\'{ITEM}\'".format(ITEM=item) for item in items]
    return ",".join(items)

def encode_2Dlist_as_row_list_of_strings(items1, items2):
    """
        Encodes a 2 lists as a row list of items for Athena query.
        
        Adding a list of items to athena query argument requires it to be
        encoded as a string encoded list
    """
    all_combinations = []
    list1_permutations  = itertools.permutations(items1, len(items2))
    for each_permutation in list1_permutations:
        zipped = zip(each_permutation, items2)
        all_combinations += list(zipped)
    # print(all_combinations)
    items = ["ROW(\'{ITEM1}\', \'{ITEM2}\')".format(ITEM1=combo[0],ITEM2=combo[1]) for combo in all_combinations]
    return ",".join(items)



# -*- coding: utf-8 -*-
"""
Creating features for analysis
"""

import pandas as pd
from sklearn.preprocessing import OneHotEncoder
from sklearn.impute import SimpleImputer


def data_processing(data: pd.DataFrame, locations: pd.DataFrame):
    """ takes in specific dataset and creates new columns/features for analysis and modeling

    note:

        - can take in geo locations in specific format
        -  joins the 2 tables and then creates various new columns

    Args:
        data (pd.DataFrame): pandas dataframe with the following columns: 'listing_id', 'user_id', 'target', 'max_level_experience',
               'years_total', 'skills_count', 'max_level_qualification', 'location',
               'country', 'industries', 'categories', 'salary_expectation',
               'qualification_types', 'min_level_experience', 'min_years',
               'min_level_qualification', 'ls_location', 'ls_country', 'l_industry',
               'l_job_function', 'salary_min', 'salary_max', 'qualification_levels'

    Returns:
        featureset (pd.DataFrame): a pandas dataframe with the features ready for analysis and model processing


    """
    geo_data = pd.merge(
        data[["user_id", "location", "country", "ls_location", "ls_country"]],
        locations,
        how="left",
        left_on=["location", "country"],
        right_on=["location", "country"],
        suffixes=("_c", "_cgeo"),
    )
    geo_data = pd.merge(
        geo_data,
        locations,
        how="left",
        left_on=["ls_location", "ls_country"],
        right_on=["location", "country"],
        suffixes=("_c", "_l"),
    )

    geo_data = geo_data[["lat_c", "lon_c", "lat_l", "lon_l"]]
    geo_data["dist"] = np.nan
    for i, r in geo_data.iterrows():
        x1 = float(r["lat_c"])
        x2 = float(r["lat_l"])
        y1 = float(r["lon_c"])
        y2 = float(r["lon_l"])

        dist = math.sqrt((x2 - x1) ** 2 + (y2 - y1) ** 2)
        geo_data["dist"][i] = dist
    # add flags for missing loc req and for missing loc c

    # get ratios between columns
    col = "max_level_experience"
    col = "years_total"
    col = "skills_count"
    col = "max_level_qualification"

    ratios = pd.DataFrame(data["user_id"])

    # get non empty rows
    ratios["exp_ratio"] = (
        data["max_level_experience"].dropna().div(data["min_level_experience"].dropna())
    )
    ratios["qual_ratio"] = (
        data["max_level_qualification"]
        .dropna()
        .div(data["min_level_qualification"].dropna())
    )
    ratios["years_ratio"] = data["years_total"].dropna().div(data["min_years"].dropna())

    industries = pd.DataFrame(data["user_id"])
    industries["industry_match"] = np.nan

    industries["function_match"] = np.nan
    # match industries
    for i, r in data.iterrows():
        if r["industries"] is not None:
            can_ind = r["industries"].split(",")
            list_ind = r["l_industry"]
            # find if listing industry is in the candidate list
            if list_ind in can_ind:
                industries["industry_match"][i] = 1
            else:
                industries["industry_match"][i] = 0

        if r["categories"] is not None:
            can_func = r["categories"].split(",")
            list_func = r["l_job_function"]
            # find if listing industry is in the candidate list
            if list_func in can_func:
                industries["function_match"][i] = 1
            else:
                industries["function_match"][i] = 0

    data[["l_industry", "l_job_function"]] = data[
        ["l_industry", "l_job_function"]
    ].fillna("Other")

    # convert categories to features
    enc = OneHotEncoder(handle_unknown="ignore")
    enc.fit(data[["l_industry", "l_job_function"]])
    # remember to output the encoder as a loblib before deploying
    categorical = pd.DataFrame(
        enc.transform(data[["l_industry", "l_job_function"]]).toarray(),
        columns=enc.get_feature_names(),
    )

    # salary between range
    salary = pd.DataFrame(data["user_id"])
    salary["salary_match"] = np.where(
        (data["salary_min"] <= data["salary_expectation"])
        & (data["salary_max"] >= data["salary_expectation"]),
        1,
        0,
    )
    # salary expect % of max
    salary["salary_ratio"] = data["salary_expectation"].div(data["salary_max"])

    # qualification match
    # no data for this now
    # fuzzy match percentage between- will work if application and lisitng are in the same language

    # fuzz.token_sort_ratio('To be or not to be', 'To be not or to be')

    # Combine into final feature set

    featureset = data[["listing_id", "user_id", "target"]]

    featureset["location_distince"] = geo_data["dist"]
    featureset["missing_location"] = geo_data["dist"]
    # replace nan with mean value and give missing value indicator
    geo_imputer = SimpleImputer(missing_values=np.nan, strategy="mean")
    geo_imputer = geo_imputer.fit(
        np.array(featureset["location_distince"]).reshape(-1, 1)
    )
    a = featureset["missing_location"]
    a[~np.isnan(a)] = 0
    a[np.isnan(a)] = 1
    featureset["location_distince"] = geo_imputer.transform(
        np.array(featureset["location_distince"]).reshape(-1, 1)
    )

    # add ratios
    ratios.replace([np.inf, -np.inf], np.nan, inplace=True)
    featureset[["exp_ratio", "qual_ratio", "years_ratio"]] = ratios[
        ["exp_ratio", "qual_ratio", "years_ratio"]
    ]
    featureset[
        ["missing_exp_ratio", "missing_qual_ratio", "missing_years_ratio"]
    ] = ratios[["exp_ratio", "qual_ratio", "years_ratio"]]

    featureset["missing_exp_ratio"][~np.isnan(featureset["missing_exp_ratio"])] = 0
    featureset["missing_exp_ratio"][np.isnan(featureset["missing_exp_ratio"])] = 1

    featureset["missing_qual_ratio"][~np.isnan(featureset["missing_qual_ratio"])] = 0
    featureset["missing_qual_ratio"][np.isnan(featureset["missing_qual_ratio"])] = 1

    featureset["missing_years_ratio"][~np.isnan(featureset["missing_years_ratio"])] = 0
    featureset["missing_years_ratio"][np.isnan(featureset["missing_years_ratio"])] = 1

    geo_imputer = SimpleImputer(missing_values=np.nan, strategy="mean")
    geo_imputer = geo_imputer.fit(
        featureset[["exp_ratio", "qual_ratio", "years_ratio"]]
    )
    featureset[["exp_ratio", "qual_ratio", "years_ratio"]] = geo_imputer.transform(
        featureset[["exp_ratio", "qual_ratio", "years_ratio"]]
    )

    featureset[["industry_match", "function_match"]] = industries[
        ["industry_match", "function_match"]
    ]
    featureset[["missing_industry", "missing_function"]] = industries[
        ["industry_match", "function_match"]
    ]

    featureset["missing_industry"][~np.isnan(featureset["missing_industry"])] = 0
    featureset["missing_industry"][np.isnan(featureset["missing_industry"])] = 1
    featureset["missing_function"][~np.isnan(featureset["missing_function"])] = 0
    featureset["missing_function"][np.isnan(featureset["missing_function"])] = 1

    geo_imputer = SimpleImputer(missing_values=np.nan, strategy="mean")
    geo_imputer = geo_imputer.fit(featureset[["industry_match", "function_match"]])
    featureset[["industry_match", "function_match"]] = geo_imputer.transform(
        featureset[["industry_match", "function_match"]]
    )

    salary.replace([np.inf, -np.inf], np.nan, inplace=True)
    featureset[["salary_ratio", "salary_match"]] = salary[
        ["salary_ratio", "salary_match"]
    ]
    featureset["salary_ratio"] = np.where(
        featureset["salary_ratio"] > 5, 6, featureset["salary_ratio"]
    )
    featureset[["missing_salary_ratio", "missing_salary_match"]] = salary[
        ["salary_ratio", "salary_match"]
    ]

    featureset["missing_salary_ratio"][
        ~np.isnan(featureset["missing_salary_ratio"])
    ] = 0
    featureset["missing_salary_ratio"][np.isnan(featureset["missing_salary_ratio"])] = 1
    featureset["missing_salary_match"][
        ~np.isnan(featureset["missing_salary_match"])
    ] = 0
    featureset["missing_salary_match"][np.isnan(featureset["missing_salary_match"])] = 1

    featureset=featureset.join(categorical)#this increases the column qty, should be removed if not used for ML,
    featureset.fillna(0, inplace=True)

    return featureset



"""
    distance scoring methods

    methods include
    ------------------------------------
    1. adjusted_cosine
    2. score_method_handler
"""
from scipy import spatial
import numpy as np

def adjusted_cosine(base_vector, comp_vector):
    """finds indices in entity_vector that equal, removes in both vectors and applies cosine

    Args:
        base_vector (list): base vector
        comp_vector (list): comparison vector

    Returns:
        float: adjusted cosine score
    """
    if isinstance(base_vector, np.ndarray):
        base_vector = base_vector.tolist()

    if isinstance(comp_vector, np.ndarray):
        comp_vector = comp_vector.tolist()

    indices = [
        i for i, x in enumerate(base_vector) if x == 0
    ]  # must be from joblisting

    indices = sorted(indices, reverse=True)

    for idx in indices:
        comp_vector.pop(idx)

    for idx in indices:
        base_vector.pop(idx)

    return 1 - spatial.distance.cosine(base_vector, comp_vector)

def score_method_handler(base_vector: list, comp_vector: list, scorer_types: list) -> dict:
    """scorer method handler

    Args:
        base_vector (list): base vector
        comp_vector (list): comparison vector
        scorer_types (list): scorer types to run scoring for

    Raises:
        TypeError: _description_

    Returns:
        dict: scores for given vectors
    """
    scores = {}

    if "simple_cosine" in scorer_types:

        score = 1 - spatial.distance.cosine(base_vector, comp_vector)
        scores["simple_cosine"] = score

    if "adjusted_cosine" in scorer_types:

        scores["adjusted_cosine"] = adjusted_cosine(base_vector, comp_vector)

    if "simple_euclidean" in scorer_types:

        # score = np.linalg.norm(np.asarray(base_1)-np.asarray(comp_1))#this is theoratically faster function
        score = 1 / (
            1 + spatial.distance.euclidean(base_vector, comp_vector)
        )  # this is a slower function, should actually use np sqrt
        scores["simple_euclidean"] = score

    if len(scores.keys()) == 0:
        raise TypeError(f"None of `scorer_types` == {scorer_types} are currently supported")

    return scores



"""
    jarvis service utilities

    methods include
    ------------------------
    1. read_file_stream_from_s3
    2. compile_document_for_request
    3. fetch_document_vectors_from_api
    4. slice_baskets
    5. compile_documents_for_jarvis
    6. batch_and_fetch_vectors

"""
import base64
import io
import math
import requests
from typing import List, Any, Dict
import boto3

def read_file_stream_from_s3(bucket_name: str, file_path: str):
    """read file from s3 as a file_stream ready for json encoding

    Args:
        bucket_name (str): name of s3 bucket
        file_path (str): s3 prefix of full filepath

    Returns:
        str: json supported file stream
    """
    s3_resource = boto3.resource('s3')
    bucket = s3_resource.Bucket(bucket_name)

    object = bucket.Object(file_path)
    file_stream = io.BytesIO()
    object.download_fileobj(file_stream)
    file_stream.seek(0)

    file_stream = file_stream.getvalue()
    return str(base64.b64encode(file_stream))[2:-1]

def compile_document_for_request(
    file_stream: str,
    doc_type: str,
    file_type: str,
    unique_id: int,
) -> dict:
    """compile document in to a jarvis supported dictionary document

    Args:
        file_stream (str): json supported file stream
        doc_type (str): jarvis supported document type i.e. job or cv
        file_type (str): document type i.e. pdf, txt, etc
        unique_id (int): unique entity id

    Raises:
        TypeError: document type not supported

    Returns:
        dict: jarvis supported document
    """
    file_type = file_type.lower()
    if (
        file_type == "doc"
        or file_type == "docx"
        or file_type == "pdf"
        or file_type == "txt"
        or file_type == "html"
        or file_type == "rtf"
    ):

        document_dict = {
            "id": str(unique_id),
            "doctype": doc_type,
            "base64Data": file_stream,
            "filetype": str(file_type)
        }
        return document_dict

    else:
        raise TypeError(f"Error on document type = {file_type} is not supported")

def fetch_document_vectors_from_api(documents: list, embedding_url: str, api_key: str):
    """fetch vectors for given list of documents from jarvis

    Args:
        documents (list): _description_
        embedding_url (str): _description_
        api_key (str): _description_

    Raises:
        requests.exceptions.RequestException: _description_

    Returns:
        list, list: records, errors
    """
    documents_json = {"documents": documents}

    header = {
        "Authorization": "Bearer " + str(api_key),
        "Content-Type": "application/json"
    }

    response = requests.post(embedding_url, headers=header, json=documents_json)

    if response.status_code != 200:
        # should the lambda fail on error? Or just write error to athena?
        raise requests.exceptions.RequestException(
            "error ({}): {}".format(response.status_code, response.raise_for_status())
        )


    resp_json = response.json()

    results = []
    if "results" in resp_json:
        # structure results for response
        
        for record in resp_json["results"]:
            result = {}

            result["entity_id"] = record["id"]
            result["properties"] = {}
            result["properties"]["doctype"] = record["doctype"]
            result["properties"]["title"] = record["title"]
            result["properties"]["skills"] = record["skills"]

            result["properties"]["embedding"] = {}
            result["properties"]["embedding"]["dim"] = record["embedding"]["dim"]
            result["properties"]["embedding"]["vector"] = record["embedding"]["vector"]
            results.append(result)

    errors = []
    if "errors" in resp_json:
        # individual items can fail due to API not being able to extract information or match skills
        
        for record in resp_json["errors"]:
            error = {}
            error["errorCode"] = record["errorCode"]
            error["errorCode"] = record["errorMessage"]
            # structure errors to raise
            errors.append(error)

    return results, errors


def slice_list_in_to_buckets(items: List[Any], max_buckets: int = None, chunk_size: int = None) -> List[List[Any]]:
    """create bucketted list of items from a single list of items

    Args:
        items (List[Any]): items to bucket
        max_buckets (int, optional): max number of buckets. Defaults to None.
        chunk_size (int, optional): max size of each bucket. Defaults to None.

    Raises:
        Exception: Must provide max_buckets or chunk_size

    Returns:
        List[List[Any]]: bucketted list of items
    """
    if chunk_size is not None:
        max_buckets = math.ceil(len(items) / chunk_size)
    if max_buckets is None:
        raise Exception("Provide either max_buckets or chunk_size.")
    n_baskets = min(max_buckets, len(items))
    return [items[i::n_baskets] for i in range(n_baskets)]

def compile_documents_for_jarvis(records: List[Dict[str, str]]) -> List[Dict[str, str]]:
    """compile a document records for jarvis

    Args:
        records (List[Dict[str, str]]): records to compile for jarvis

    Returns:
        List[Dict[str, str]]: compiled documents for jarvis
    """
    compiled_documents = []

    for record in records:
        entity_id = record["entity_id"]
        entity_type = record["entity_type"]
        bucket_name = record["bucket"]
        file_path = record["file_path"]
        file_type = record["file_type"]

        # execute scoring workflow
        file_stream = read_file_stream_from_s3(bucket_name, file_path)
        compile_document = compile_document_for_request(
            file_stream=file_stream,
            doc_type="cv" if entity_type == "candidates" else "job",
            file_type=file_type,
            unique_id=entity_id
        )
        compiled_documents.append(compile_document)
    
    return compiled_documents

def batch_and_fetch_vectors(records: List[Dict[str, str]], compiled_documents: List[Dict[str, str]], embedding_url: str, api_key: str, **kwargs):
    """batch reocrds and fetch vectors from jarvis

    Args:
        records (List[Dict[str, str]]): document records
        compiled_documents (List[Dict[str, str]]): compiled document records
        embedding_url (str): url for jarvis api
        api_key (str): api key for jarvis api

    Returns:
        list, list: results, errors
    """
    created_at = kwargs.get('created_at')
    chunk_size = kwargs.get('chunk_size', 10)

    print("batch and fetch documents from jarvis")
    final_results = []
    final_errors = []
    chunk_index = slice_list_in_to_buckets(items=range(len(compiled_documents)), chunk_size=chunk_size)
    for chunk in chunk_index:

        records_i = [records[i] for i in chunk]
        compiled_documents_i = [compiled_documents[i] for i in chunk]

        results, errors = fetch_document_vectors_from_api(
            documents=compiled_documents_i,
            embedding_url=embedding_url,
            api_key=api_key
        )

        for i, record_i in enumerate(records_i):
            results[i].update(
                {
                    'entity_type': record_i["entity_type"],
                    'source': 'jarvis',
                    'created_at': created_at
                }
            )

        final_results.extend(results)
        final_errors.extend(errors)
    
    return final_results, final_errors



"""
    database processing utilities

    methods include
    --------------------
    1. convert_listing_to_document
    2. convert_candidate_to_document
    3. find_structured_data_and_create_file
"""
from typing import List, Dict
from pathlib import Path
import re
from dp_tms.database.storage import \
    check_if_cloud_storage_and_extract_path, write_txt_to_s3

def convert_listing_to_document(data: dict) -> str:
    """converts a structured listing json payload in to a txt supported text dump
    Args:
        data (dict): structured json/dict payload
    Returns:
        (str): text dump
    """
    listing_title = re.sub(r"[^a-zA-Z0-9]", " ", data["title"])
    industry = re.sub(r"[^a-zA-Z0-9]", " ", data["industry"])
    experience_lengths = re.sub(
        r"[^a-zA-Z0-9]", " ", data["experience_lengths"]
    )
    experience_levels = re.sub(
        r"[^a-zA-Z0-9]", " ", data.get("experience_levels", "")
    )
    qualification_levels = re.sub(
        r"[^a-zA-Z0-9]", " ", data.get("qualification_levels", "")
    )
    location = re.sub(r"[^a-zA-Z0-9]", " ", data["location"])
    description = re.sub(r"[^a-zA-Z0-9]", " ", data["description"])
    details = ""
    details = re.sub("<[^<]+?>", "", data.get("details", ""))
    details = re.sub(r"[^a-zA-Z0-9]", " ", details)

    # combine to create listing document
    text = (
        """Title \r\n """
        + listing_title
        + """  \r\n\r\n 
            Description \r\n"""
        + description
        + """ \r\n\r\n 
            Industry \r\n"""
        + industry
        + """\r\n\r\n        
            Experience level \r\n"""
        + experience_levels
        + """\r\n\r\n
            Experience length \r\n"""
        + experience_lengths
        + """\r\n\r\n
            location \r\n"""
        + location
        + """\r\n\r\n
            Qualification \r\n"""
        + qualification_levels
        + """\r\n\r\n
            Details \r\n"""
        + details
    )

    return text

def convert_candidate_to_document(data: dict) -> str:
    """converts a structured candidate json payload in to a txt supported text dump
    Args:
        data (dict): structured json/dict payload
    Returns:
        (str): text dump
    """
    headline = re.sub(r"[^a-zA-Z0-9]", " ", data["headline"])
    qualification = re.sub(r"[^a-zA-Z0-9]", " ", data["qualification"])
    skills = re.sub(r"[^a-zA-Z0-9]", " ", data["skills"])
    languages = re.sub(r"[^a-zA-Z0-9]", " ", data["languages"])
    experience = re.sub(r"[^a-zA-Z0-9]", " ", data["experience"])
    cover_letter = re.sub(r"[^a-zA-Z0-9]", " ", data["cover_letter"])

    # convert to CV format in txt
    text = (
        """CV \r\n """
        + headline
        + """  \r\n\r\n 
            Cover Letter \r\n"""
        + cover_letter
        + """ \r\n\r\n 
    
            Work Experience  \r\n"""
        + experience
        + """\r\n\r\n
            Qualifications \r\n"""
        + qualification
        + """\r\n\r\n
            Skills \r\n"""
        + skills
        + """\r\n\r\n
            Language Skills \r\n"""
        + languages
        + """\r\n\r\n"""
    )
    return text

def find_structured_data_and_create_file(entities: List[Dict[str, str]], entity_type: str, base_storage_path: str, created_at: str) -> List[Dict[str, str]]:
    """
    finds if resume/document path exists,
    else attempts to construct a resume/document document from structured json
    and write a file to storage location.
    Keeps a reference in processed entities.

    Args:
        entities (List[Dict[str, str]]): entity records
        entity_type (str): type of entities being processed
        base_storage_path (str): path that file storage paths are based on
        created_at (str): date at which entities are processed and staged

    Returns:
        List[Dict[str, str]]: processed entity records
    """
    entities = entities.copy()
    for i, entity in enumerate(entities):
        # iterate and check
        entity_id = entity.get("entity_id")
        properties = entity.get("properties", {})
        resume = properties.get("resume")
        if resume is None:

            if entity_type == "listings":

                # create document
                content = convert_listing_to_document(properties)
            elif entity_type == "candidates":
                content = convert_candidate_to_document(properties)

            storage_path = f"{base_storage_path}/{entity_id}.txt"
            print(f"store at: {storage_path}")

            write_txt_to_s3(storage_path, content)


            bucket, file_path, scheme = check_if_cloud_storage_and_extract_path(storage_path, schemes = ['s3'])

            storage_reference = {
                "file": file_path,
                "type": "txt",
                "disk": scheme,
                "bucket": bucket
            }

        else:
            file_type = resume.get("type")
            if file_type is None:
                file_path = resume["file"]
                file_type = Path(file_path).suffixes
                if len(file_type) > 0:
                    file_type = "".join(file_type)
                else:
                    file_type = None

            storage_reference = {
                "file": resume["file"],
                "type": file_type,
                "disk": resume["disk"],
                "bucket": resume["bucket"]
            }

            storage_reference = resume

        if entity_type == "listings":
            # create document
            entities[i]["entity_type"] = entities[i].get("entity_type", "listing")
        elif entity_type == "candidates":
            entities[i]["entity_type"] = entities[i].get("entity_type", "candidate")

        entities[i]["properties"]["storage_reference"] = storage_reference
        entities[i]["created_at"] = created_at

    return entities
